
\ Forth compiler for MTV (Much Tinier VM)
\
\ Robert Oestling
\ robost@telia.com
\ 2003-01-06

create buffer $4000 allot
buffer var buf-ptr
$68    var ram-ptr
0      var current

$00 const i_drop
$01 const i_dup
$02 const i_over
$03 const i_swap
$04 const i_4drop
$05 const i_4dup
$06 const i_4over
$07 const i_4swap
$08 const i_adrop
$09 const i_adup
$0a const i_aover
$0b const i_aswap
$0c const i_+
$0d const i_-
$0e const i_and
$0f const i_or
$10 const i_xor
$11 const i_not
$12 const i_neg
$13 const i_2*
$14 const i_2/
$15 const i_4+1
$16 const i_4-1
$17 const i_4neg
$18 const i_<<
$19 const i_>>
$1a const i_4+4
$1b const i_4-4
$1c const i_4and
$1d const i_4or
$1e const i_4xor
$1f const i_=
$20 const i_<>
$21 const i_<
$22 const i_>
$23 const i_4=1
$24 const i_4=4
$25 const i_4=0
$26 const i_4<0
$27 const i_4>=0
$28 const i_c@
$29 const i_c!
$2a const i_c@+
$2b const i_c!+
$2c const i_@
$2d const i_!
$2e const i_>r
$2f const i_r>
$30 const i_>a
$31 const i_a>
$32 const i_if
$33 const i_if0
$34 const i_is
$35 const i_is0
$36 const i_jump
$37 const i_jumpf
$38 const i_jump8
$39 const i_call
$3a const i_callf
$3b const i_call8
$3c const i_ret
$3c const i_;
$3d const i_retf
$3e const i_lit
$3e const i_#
$3f const i_4lit
$3f const i_4#
$40 const i_alit
$40 const i_a#
$41 const i_ia#
$42 const i_sys

defer IF    ' if   is IF
defer ELSE  ' else is ELSE
defer THEN  ' then is THEN

defer IS  ' is is IS

: output   buf-ptr c!  1 +!> buf-ptr ;
: output4  buf-ptr !   4 +!> buf-ptr ;
: single  create c,  does> c@ output ;
: param   create c,  does> c@ output  output4 ;
: param8  create c,  does> c@ output  output ;

: here  buf-ptr buffer - ;
: dump  buffer  here  for dup c@ h. 1+ nxt  drop ;
: error  ." Error: can not create file!" cr ;
: write  >r  here  buffer  r@ <write> drop  r> <close> drop ;
: save  $1ff bl word  hhere count s>z <creat> dup 0> if write exit then error ;

: store-count  2dup swap c! swap 1+ swap ;
: strcpy  store-count  for over c@ over c! 1+ swap 1+ swap nxt 2drop ;
: "  $22 parse dup >r  buf-ptr swap  strcpy  r> 1+ +!>  buf-ptr ;
: check-rel  here 1+ swap -  abs 126 > if ." Relative jump too long." cr then ;
: rel-jump  dup check-rel  here 1+ -  $ff and  output ;
: rel-patch  dup buffer - check-rel  here over buffer - - 1-  swap c! ;
: compile-patch  here swap ! ;
: call-rel  i_call8 output  rel-jump ;
: call-abs  i_call output  output4 ;
: compile-call  dup here 2+ - abs 127 < if call-rel exit then call-abs ;
: jump-rel  i_jump8 output  rel-jump ;
: jump-abs  i_jump output  output4 ;
: compile-jump  dup here 2+ - abs 127 < if jump-rel exit then jump-abs ;
: goto  ' >body @  compile-jump ;
: {  here ;
: }  compile-jump ;
: label  here const ;
: if   i_if0 output  buf-ptr  0 output ;
: if0  i_if  output  buf-ptr  0 output ;
: is   i_is0 output  buf-ptr  0 output ;
: is0  i_is  output  buf-ptr  0 output ;
: start  i_jump output  buf-ptr  0 output4 ;
: entry:  compile-patch ;
: then  rel-patch ;
: tail  i_jump output  current output4 ;
: constant   create c,  does> c@  i_#  output  output ;
: 4constant  create ,  does> @    i_4# output  output4 ;
: allot  +!> ram-ptr ;
: variable   create ram-ptr ,  1 allot  does> @  i_a# output  output4 ;
: 4variable  create ram-ptr ,  4 allot  does> @  i_a# output  output4 ;
: syscall  create c,  does> c@  i_# output  output  i_sys output ;

: : create  here !> current  here ,  does> @  compile-call ;

defer ,   ' output  IS , ;
defer 4,  ' output4 IS 4, ;

i_drop single drop
i_dup single dup
i_over single over
i_swap single swap
i_4drop single 4drop
i_4dup single 4dup
i_4over single 4over
i_4swap single 4swap
i_adrop single adrop
i_adup single adup
i_aover single aover
i_aswap single aswap
i_+ single +
i_- single -
i_and single and
i_or single or
i_xor single xor
i_not single not
i_neg single neg
i_2* single 2*
i_2/ single 2/
i_4+1 single 4+1
i_4-1 single 4-1
i_4neg single 4neg
i_<< single <<
i_>> single >>
i_4+4 single 4+4
i_4-4 single 4-4
i_4and single 4and
i_4or single 4or
i_4xor single 4xor
i_= single =
i_<> single <>
i_< single <
i_> single >
i_4=1 single 4=1
i_4=4 single 4=4
i_4=0 single 4=0
i_4<0 single 4<0
i_4>=0 single 4>=0
i_c@ single c@
i_c! single c!
i_c@+ single c@+
i_c!+ single c!+
i_@ single @
i_! single !
i_>r single >r
i_r> single r>
i_>a single >a
i_a> single a>
i_ret single ret
i_; single ;
i_retf single retf
i_sys single sys

i_jump param jump
i_jumpf param jumpf
i_call param call
i_callf param callf
i_4lit param 4lit
i_4# param 4#
i_alit param alit
i_a# param a#
i_ia# param ia#

i_lit param8 lit
i_# param8 #
i_jump8 param8 jump8
i_call8 param8 call8

