\ mtvfc - MTV Forth Compiler
\
\ Robert Oestling
\ robost@telia.com (robost86@hotmail.com)
\ http://robert.zizi.org/
\
\ Started 2003-04-30
\ Some minor fixes 2003-06-10
\
\ Written for IsForth 1.13b (http://isforth.clss.net)
\ Thanks to I440r for writing this neat Linux Forth :)
\
\ The old Forth compiler for MTV was just a dirty hack, this time I
\ intend to make it good. Since it replaced many basic Forth words
\ and because it's a cross compiler, you couldn't use any compile-time
\ words. And, uh.. what's the point of using Forth then?
\
\ Word header.
\
\ 00  dword  Pointer to previous word header.
\ 04  dword  Pointer to the corresponding IsForth word to run.
\ 08  byte   Name length.
\ 09         Name string.

fload opcodes.f

$68 const ram-start

\ FUCK! Don't EVER make this one too short. Took me hours of debugging.
create buff  $10000 allot

buff      var buffer
buffer    var prog-ptr
0         var mtv-last
ram-start var ram-ptr


false var mtv-compile?

: offset ( -- n )  prog-ptr buffer - ;

: compile-1 ( n -- )  prog-ptr c!  1 +!> prog-ptr ;
: compile-4 ( n -- )  prog-ptr !   4 +!> prog-ptr ;

: compile-lit1 ( n -- )  i_lit compile-1 compile-1 ;
: compile-lit4 ( n -- )  i_4lit compile-1 compile-4 ;
: compile-lit ( n -- )
  dup $100 <
  if  goto compile-lit1  then          \ If below 2^8, compile a lit.
  compile-lit4 ;                       \ Else, compile a 4lit.

: compile-alit4 ( n -- )
  i_alit compile-1 compile-4 ;

' compile-alit4 alias compile-alit

\ Ooops! You can't do it like this... I'm so stupid.
\
\ : compile-alit ( n -- )
\   dup $100 <                           \ If below 2^8, compile lit >a
\   if  compile-lit1 i_>a compile-1 exit  then
\   compile-alit4 ;                      \ Otherwise, compile alit.

\ Here we have to remember that 1-byte addresses are relative, while 4-byte
\ addresses are absolute.

: compile-reljump ( offset -- )
  offset 1+ -  $ff and  compile-1 ;

: compile-call1 ( offset -- )
  i_call8 compile-1  compile-reljump ;
: compile-call4 ( offset -- )
  i_call compile-1  compile-4 ;
: compile-call ( offset -- )
  dup offset - abs 126 < if  goto compile-call1  then compile-call4 ;

: compile-jump1 ( offset -- )
  i_jump8 compile-1  compile-reljump ;
: compile-jump4 ( offset -- )
  i_jump compile-1  compile-4 ;
: compile-jump ( offset -- )
  dup offset - abs 126 < if  goto compile-jump1  then compile-jump4 ;


: compare-pair ( a1 a2 -- a1 a2 not-equal? )
  2dup c@ swap c@ <>                   \ Check if the bytes are inequal.
  >r 1+ swap 1+ swap r> ;              \ Add 1 to both addresses.

: s= ( a1 a2 -- n )
  dup c@ 1+ for                        \ Get length of string 2.
    compare-pair                       \ Check if bytes are inequal.
    if  2drop false r>drop exit  then  \ Yes? Then push false and quit.
  nxt 2drop true ;                     \ Equal - push yes.

: lookup ( a1 -- a | false )
  mtv-last begin                       \ Pointer to last word in list.
    dup 0<>                            \ Null? Then break the loop.
  while
    2dup 4+ 4+ s= if  nip exit  then   \ Is this what we want? Yes - exit.
  @ repeat                             \ No, get next item of list.
  2drop false ;                        \ We didn't find anything, ret false.

: word" ( a1 -- )
  here mtv-last , !> mtv-last          \ Copy previous word header, update it.
  , ," ;                               \ Copy word address and name.

: read ( -- a )  bl word ;             \ Read word. hhere returns addr to it.

: run ( a1 -- success? )
  lookup dup 0= if  exit  then         \ Return if word doesn't exist.
  4+ @ execute  true ;                 \ Run and return true.

: run-isforth ( -- success? )
  find 0= if  false exit  then         \ If word isn't defined, return false.
  execute true ;                       \ Otherwise, execute word.
  
: run ( -- success? )
  hhere lookup ?dup 0=                 \ Look up the word.
  if  goto run-isforth  then           \ If it's not there, let isforth try.
  4+ @  execute  true ;

: error-msg ( -- )
  hhere count type  ."  <-- unknown word." cr ;

: read-number ( -- n )
  hhere number 0= if  error-msg 0 exit  then ;

\ It's important to check for a number first here.
: process ( -- )
  hhere c@ 0 = if exit then            \ Don't bother about empty words.
  hhere 1+ c@ '^' = if                 \ First char = '^' ?
    hhere c@ 1- hhere 1+ c!            \ Yes, remove first char from string.
    hhere 1+ number 0= if              \ A number?
      error-msg                        \ No, print error message.
    then exit
  then
  hhere number                         \ Convert string to integer.
  0= if                                \ Invalid number?
    run not if                         \ Yes, try to run word.
      error-msg exit                   \ Error? Yes, print error msg and exit.
    then exit                          \ Quit if the word was run.
  then compile-lit ;                   \ Number? Compile literal.

: [[ ( -- )  false !> mtv-compile? ;   \ Turn compilation off.
' [[ word" [["

: ]] ( -- )
  true !> mtv-compile?                 \ Turn compilation on.
  begin
    read process                       \ Read one word and process it.
  mtv-compile? not until ;             \ Repeat until [[ is run.

: single ( opcode -- )
  create c,  does> c@ compile-1 ;      \ When word is run, compile opcode.

: param ( opcode -- )
  create c,  does> c@ compile-1        \ When word is run, compile opcode...
  compile-4 ;                          \ ...and the parameter.

: param8 ( opcode -- )
  create c,  does> c@ compile-1        \ Same as above, but with 8-bit
  compile-1 ;                          \ parameter.


: patch-rel ( a -- )
  dup offset 1+ - abs  126 >
  if
    ." Relative jump too long." cr drop exit
  then
  offset over - 1-  swap buffer + c! ;

: patch-abs ( a -- )
  offset swap  buffer + ! ;

: mtv-:  create  offset ,  does> @ compile-call ;
' mtv-: word" :"
: mtv-' ' >body @ ;
' mtv-' word" '"
: mtv-4#   read read-number  i_4# compile-1  compile-4 ;
' mtv-4# word" 4#"
: mtv-a#   read read-number  i_a# compile-1  compile-4 ;
' mtv-a# word" a#"
: mtv-ia#  read read-number  i_ia# compile-1  compile-4 ;
' mtv-ia# word" ia#"
: mtv-goto  ' >body @  compile-jump ;
' mtv-goto word" goto"
: mtv-label  offset const ;
' mtv-label word" label"
: mtv-create  create offset , does> @ i_ialit compile-1 compile-4 ;
' mtv-create word" create"
: mtv-ram-create  create ram-ptr ,  does> @ i_alit compile-1 compile-4 ;
' mtv-ram-create word" ram-create"

: mtv-if  i_if0 compile-1  offset 0 compile-1 ;
' mtv-if word" if"
: mtv-if0  i_if compile-1  offset 0 compile-1 ;
' mtv-if0 word" if0"
: mtv-is  i_is0 compile-1  offset 0 compile-1 ;
' mtv-is word" is"
: mtv-is0  i_is compile-1  offset 0 compile-1 ;
' mtv-is0 word" is0"
: mtv-then  patch-rel ;
' mtv-then word" then"

: ram-allot ( n -- )  +!> ram-ptr ;

: compile-var1 ( a -- )                \ 5 or 7 bytes.
  compile-alit                         \ "alit a"
  i_c@ compile-1                       \ c@
  i_adrop compile-1 ;                  \ adrop
: compile-var4 ( a -- )                \ 5 or 7 bytes.
  compile-alit                         \ "alit a"
  i_@ compile-1                        \ @
  i_adrop compile-1 ;                  \ adrop

: mtv-var ( -- )
  ram-ptr  1 ram-allot                 \ Get RAM ptr, alloc. 1 byte of RAM.
  create 0 c, ,                        \ Create word with pointer in body.
  does> 1+ @ compile-var1 ;
' mtv-var word" var"

: mtv-var4 ( -- )
  ram-ptr  4 ram-allot                 \ Get RAM ptr, alloc. 4 bytes of RAM.
  create 1 c, ,                        \ Create word with pointer in body.
  does> 1+ @ compile-var4 ;
' mtv-var4 word" var4"

: compile-to1 ( a -- )
  compile-alit i_c! compile-1          \ "alit a  c!"
  i_adrop compile-1 ;                  \ adrop
: compile-to4 ( a -- )
  compile-alit i_! compile-1           \ "alit a  !"
  i_adrop compile-1 ;                  \ adrop

: mtv--> ( -- )
  ' >body dup c@ 0=                    \ 1-byte var?
  if  1+ @ compile-to1 exit  then      \ Yes, compile the var1 code and exit.
  1+ @ compile-to4 ;                   \ Otherwise, compile the var4 code.
' mtv--> word" ->"

: mtv-address ( -- )                   \ Use: var a  address a c@
  ' >body 1+ @  compile-alit ;
' mtv-address word" address"

: mtv-const ( n -- )
  create ,  does> @ compile-lit ;
' mtv-const word" constant"
' mtv-const alias constant             \ Use only in IsForth mode!

: mtv-char ( -- )
  read hhere 1+ c@                     \ Read character.
  compile-lit ;                        \ Compile literal.
' mtv-char word" char"

: syscall ( n -- )
  create c,                            \ Create word, store syscall #
  does> c@  i_lit compile-1            \ Get syscall #, push it.
  compile-1                            \ Compile the syscall #
  i_sys compile-1 ;                    \ sys

\ : mtv-c, ( n -- )  compile-1 ;
\ ' mtv-c, word" c,"
\ : mtv-, ( n -- )  compile-4 ;
\  ' mtv-, word" ,"

: store-count  2dup swap c! swap 1+ swap ;
: strcpy ( src+1 dest count -- )
  store-count  for
    over c@ over c! 1+ swap 1+ swap
  nxt 2drop ;
: "  $22 parse dup >r  prog-ptr swap strcpy  r> 1+ +!> prog-ptr ;

: mtv-{ ( -- a )  offset ;
' mtv-{ word" {"
: mtv-} ( a -- )  compile-jump ;
' mtv-} word" }"

: start ( -- a )  i_jump compile-1  offset  0 compile-4 ;
: entry: ( a -- )  patch-abs ;

: sstart ( -- a )  i_jump8 compile-1  offset  0 compile-1 ;
: sentry: ( a -- )  patch-rel ;

: file-error  ." Can not create file." cr ;
: write-file ( fd -- )
  >r  offset buffer r@ <write> drop    \ Write the contents of 'buffer'
  r> <close> drop ;                    \ Close file.
: save ( -- )
  $1b6 bl word                         \ File mode:  Read filename.
  hhere count s>z                      \ Convert to ASCIIZ.
  <creat> dup 0>                       \ Create file, check for error.
  if  write-file exit  then            \ No error, write and quit.
  file-error ;

\ I guess this could be useful.
: mtv-nip ( a b -- b )
  i_swap compile-1 i_drop compile-1 ;
' mtv-nip word" nip"

\ List of words.
fload wordlist.f

