\ sim.f
\
\ MTV on AT90S8515 simulator.

\ Hack for isforth 1.13b
' << alias u<<

$200 const ramsize
$60  const ramstart

\ stack-size  stack  stack-name stack-ptr
: stack  create here over allot  + var ;

$40 stack pstack pptr
$40 stack rstack rptr
$40 stack astack aptr

$01000000 var ipbase
0         var iptr

false var halted

: init  $01000000 !> ipbase  0 !> iptr  false !> halted ;

: kb 10 u<< ;

$260    const ram-size
$200    const eeprom-size
$100 kb const flash-size

: zero ( a n -- )
  for  0 over c!  1+  nxt ;

create ram    ram-size allot
create eeprom eeprom-size allot
create flash  flash-size allot

create eeprom.bin ," eeprom.bin"
create flash.bin ," flash.bin"

0 const O_RDONLY
1 const O_WRONLY
2 const O_RDWR

: ppush  ( byte -- )  -1 +!> pptr  pptr c! ;
: ppop   ( -- byte )  pptr c@  1 +!> pptr ;
: ppush4 ( word -- )  -4 +!> pptr  pptr ! ;
: ppop4  ( -- word )  pptr @  4 +!> pptr ;
: rpush  ( word -- )  -4 +!> rptr  rptr ! ;
: rpop   ( -- word )  rptr @  4 +!> rptr ;
: apush  ( word -- )  -4 +!> aptr  aptr ! ;
: apop   ( -- word )  aptr @  4 +!> aptr ;

: 3drop 2drop drop ;

: open-file ( flags name -- fd )
  count s>z <open> ;

: read-block ( size ptr fd -- success )
  dup >r <read> 0> r> <close> drop ;

: load-block ( size ptr name -- success )
  O_RDONLY swap open-file dup 0> if read-block exit then 3drop false ;

: create-file ( mode name -- fd )
  count s>z <creat> ;

: write-block ( size ptr fd -- success )
  dup >r <write> 0> r> <close> drop ;

: save-block ( size ptr name -- success )
  \600 swap create-file dup 0> if write-block exit then 3drop false ;

: check  not if  ." Error loading from/saving to file." cr bye then ;
: load-eeprom  eeprom eeprom-size zero 
  eeprom-size eeprom eeprom.bin load-block check ;
  
: save-eeprom  eeprom-size eeprom eeprom.bin save-block check ;
: load-flash  flash flash-size zero 
  flash-size flash flash.bin load-block check ;
  
: save-flash  flash-size flash flash.bin save-block check ;
: clear-ram  ram ram-size  for 0 over c! 1+ nxt  drop ;

: load load-eeprom load-flash clear-ram ;
: save save-eeprom save-flash ;

: dummy-@ 0 ;
: dummy-! drop ;
: ram-c@ ram + c@ ;
: ram-c! ram + c! ;
: ram-@  ram + @ ;
: ram-!  ram + ! ;
: eeprom-c@ eeprom + c@ ;
: eeprom-c! eeprom + c! ;
: eeprom-@  eeprom + @ ;
: eeprom-!  eeprom + ! ;
: flash-c@ flash + c@ ;
: flash-c! flash + c! ;
: flash-@  flash + @ ;
: flash-!  flash + ! ;

create action-c@
' ram-c@ , ' eeprom-c@ , ' dummy-@ , ' flash-c@ ,
create action-c!
' ram-c! , ' eeprom-c! , ' dummy-! , ' flash-c! ,
create action-@
' ram-@ , ' eeprom-@ , ' dummy-@ , ' flash-@ ,
create action-!
' ram-! , ' eeprom-! , ' dummy-! , ' flash-! ,

: memaction  dup $00ffffff and swap 24 u>> 4* rot + @ execute ;

: mem-c@  action-c@ swap memaction ;
: mem-c!  action-c! swap memaction ;
: mem-@  action-@ swap memaction ;
: mem-!  action-! swap memaction ;

: a-c@  action-c@ aptr @ memaction ;
: a-c!  action-c! aptr @ memaction ;
: a-@  action-@ aptr @ memaction ;
: a-!  action-! aptr @ memaction ;

: op-drop   1 +!> pptr ;
: op-dup    pptr c@ ppush ;
: op-over   pptr 1+ c@ ppush ;
: op-swap   ppop ppop swap ppush ppush ;
: op-drop4  4 +!> pptr ;
: op-dup4   pptr @ ppush4 ;
: op-over4  pptr 4+ @ ppush4 ;
: op-swap4  ppop4 ppop4 swap ppush4 ppush4 ;
: op-adrop  4 +!> aptr ;
: op-adup   aptr @ apush ;
: op-aover  aptr 4+ @ apush ;
: op-aswap  apop apop swap apush apush ;

: op-+   ppop ppop + ppush ;
: op--   ppop ppop swap - ppush ;
: op-and ppop ppop and ppush ;
: op-or  ppop ppop or ppush ;
: op-xor ppop ppop xor ppush ;
: op-not ppop not ppush ;
: op-neg ppop negate ppush ;
: op-2*  ppop 2* ppush ;
: op-2/  ppop 2/ ppush ;

: op-4+1  ppop ppop4 + ppush4 ;
: op-4-1  ppop ppop4 swap - ppush4 ;
: op-4neg ppop4 negate ppush4 ;
: op-<<   ppop ppop4 swap u<< ppush4 ;
: op->>   ppop ppop4 swap u>> ppush4 ;
: op-4+4  ppop4 ppop4 + ppush4 ;
: op-4-4  ppop4 ppop4 swap - ppush4 ;
: op-4and ppop4 ppop4 and ppush4 ;
: op-4or  ppop4 ppop4 or ppush4 ;
: op-4xor ppop4 ppop4 xor ppush4 ;

: op-=  ppop ppop = ppush ;
: op-<> ppop ppop = not ppush ;
: op-<  ppop ppop > ppush ;
: op->  ppop ppop < ppush ;

: op-4=1  ppop ppop4 = ppush ;
: op-4=4  ppop4 ppop4 = ppush ;
: op-4=0  ppop4 0= ppush ;
: op-4<0  ppop4 0< ppush ;
: op-4>=0 ppop4 0< not ppush ;

: a+  1 aptr +! ;

: op-c@  a-c@ ppush ;
: op-c!  ppop a-c! ;
: op-c@+  a-c@ ppush  a+ ;
: op-c!+  ppop a-c!  a+ ;
: op-@  a-@ ppush4 ;
: op-!  ppop4 a-! ;
: op->r  ppop4 rpush ;
: op-r>  rpop ppush4 ;
: op->a  ppop4 apush ;
: op-a>  apop ppush4 ;

: b>s  24 << 24 >> ;

: get-lit  iptr ipbase + mem-c@  1 +!> iptr ;
: get-lit4  iptr ipbase + mem-@  4 +!> iptr ;

: op-if   get-lit ppop if  b>s +!> iptr exit  then drop ;
: op-if0  get-lit ppop 0= if  b>s +!> iptr exit  then drop ;
: op-is   get-lit pptr c@ if  b>s +!> iptr exit  then drop ;
: op-is0  get-lit pptr c@ 0= if  b>s +!> iptr exit  then drop ;

: op-jump   get-lit4 !> iptr ;
: op-jumpf  get-lit4 !> iptr  ppop4 !> ipbase ;
: op-jump8  get-lit b>s +!> iptr ;
: op-call   iptr 4+ rpush  op-jump ;
: op-callf  ipbase rpush  iptr 4+ rpush  op-jumpf ;
: op-call8  iptr 1+ rpush  op-jump8 ;

: op-ret   rpop !> iptr ;
: op-retf  rpop !> iptr  rpop !> ipbase ;

: op-lit   get-lit ppush ;
: op-4lit  get-lit4 ppush4 ;
: op-alit  get-lit4 apush ;
: op-ialit get-lit4 ipbase + apush ;

: sys-uart-init  ppop drop ;
: sys-uart-poll  key? ppush ;
: sys-uart-read  key ppush ;
: sys-uart-write  ppop emit ;
: sys-get-freeram  ram-size $60 - ;
: sys-sync-flash  ;
: sys-exit  true !> halted ;

create syscalls

' sys-uart-init ,
' sys-uart-poll ,
' sys-uart-read ,
' sys-uart-write ,
' sys-get-freeram ,
' sys-sync-flash ,
' sys-exit ,

: op-sys  ppop 4* syscalls + @ execute ;

create instructions

' op-drop ,
' op-dup ,
' op-over ,
' op-swap ,
' op-drop4 ,
' op-dup4 ,
' op-over4 ,
' op-swap4 ,
' op-adrop ,
' op-adup ,
' op-aover ,
' op-aswap ,
' op-+ ,
' op-- ,
' op-and ,
' op-or ,
' op-xor ,
' op-not ,
' op-neg ,
' op-2* ,
' op-2/ ,
' op-4+1 ,
' op-4-1 ,
' op-4neg ,
' op-<< ,
' op->> ,
' op-4+4 ,
' op-4-4 ,
' op-4and ,
' op-4or ,
' op-4xor ,
' op-= ,
' op-<> ,
' op-< ,
' op-> ,
' op-4=1 ,
' op-4=4 ,
' op-4=0 ,
' op-4<0 ,
' op-4>=0 ,
' op-c@ ,
' op-c! ,
' op-c@+ ,
' op-c!+ ,
' op-@ ,
' op-! ,
' op->r ,
' op-r> ,
' op->a ,
' op-a> ,
' op-if ,
' op-if0 ,
' op-is ,
' op-is0 ,
' op-jump ,
' op-jumpf ,
' op-jump8 ,
' op-call ,
' op-callf ,
' op-call8 ,
' op-ret ,
' op-retf ,
' op-lit ,
' op-4lit ,
' op-alit ,
' op-ialit ,
' op-sys ,

: exec  get-lit 4* instructions + @  execute ;
: run  init load  begin exec halted until  save ;

