; MVT - Much Tinier VM
; http://robert.zizi.org/ (UPDATE: http://robos.org/)
;
; Written by Robert Oestling
; robost@telia.com (UPDATE: robost@home.se or bob@robos.org)
; 2003-01-04 (First version)
; 2003-02-25 (DataFlash support)
;
; Temp. registers		r0:r3
; IP base			r4:r7
; IP, instruction ptr		r8:r11
; Temp. address			r12:r15
; Temp. registers		r16:r23
; AP, addressing stack ptr	r24:r25
; RP, return stack ptr		r26:r27 (X)
; PP, parameter stack ptr	r28:r29 (Y)
; Z, temporary pointer		r30:r31 (Z)


; This version of Much Tinier VM is not tiny enough to fit
; on e.g. the AT90S2313 anymore.

.include "8515def.inc"



.def	apl		= r24
.def	aph		= r25
.def	rpl		= r26
.def	rph		= r27
.def	ppl		= r28
.def	pph		= r29

.equ	ssize		= 2*6	; Hardware stack size.
.equ	asize		= 4*4	; Address stack size.
.equ	rsize		= 4*8	; Return stack size.
.equ	psize		= 32	; Parameter stack size.

.equ	stacksize	= ssize + asize + rsize + psize
.equ	freeramend	= ramend - stacksize
.equ	freeramstart	= 0x6a

.equ	buffer1l	= 0x60
.equ	buffer1h	= 0x61
.equ	buffer2l	= 0x62
.equ	buffer2h	= 0x63
.equ	bufferstat	= 0x64
.equ	bufferrecent	= 0x65

.equ	tmp1		= 0x66
.equ	tmp2		= 0x67
.equ	tmp3		= 0x68
.equ	tmp4		= 0x69

.equ	buffer1dirty	= 0
.equ	buffer2dirty	= 1


.cseg
.org 0x0000

	rjmp	reset



syscall_table:
	.dw	sys_uart_init
	.dw	sys_uart_poll
	.dw	sys_uart_read
	.dw	sys_uart_write
	.dw	sys_get_freeram
	.dw	sys_sync_flash


instruction_table:
	.dw	i_drop			; 00 drop
	.dw	i_dup			; 01 dup
	.dw	i_over			; 02 over
	.dw	i_swap			; 03 swap
	.dw	i_4drop			; 04 4drop
	.dw	i_4dup			; 05 4dup
	.dw	i_4over			; 06 4over
	.dw	i_4swap			; 07 4swap
	.dw	i_adrop			; 08 adrop
	.dw	i_adup			; 09 adup
	.dw	i_aover			; 0a aover
	.dw	i_aswap			; 0b aswap
	.dw	i_plus			; 0c +
	.dw	i_minus			; 0d -
	.dw	i_and			; 0e and
	.dw	i_or			; 0f or
	.dw	i_xor			; 10 xor
	.dw	i_not			; 11 not
	.dw	i_neg			; 12 neg
	.dw	i_mul2			; 13 2*
	.dw	i_div2			; 14 2/
	.dw	i_4plus1		; 15 4+1
	.dw	i_4minus1		; 16 4-1
	.dw	i_4neg			; 17 4neg
	.dw	i_4shl			; 18 <<
	.dw	i_4shr			; 19 >>
	.dw	i_4plus4		; 1a 4+4
	.dw	i_4minus4		; 1b 4+4
	.dw	i_4and4			; 1c 4and
	.dw	i_4or4			; 1d 4or
	.dw	i_4xor4			; 1e 4xor
	.dw	i_eq			; 1f =
	.dw	i_neq			; 20 <>
	.dw	i_be			; 21 <
	.dw	i_ab			; 22 >
	.dw	i_4eq1			; 23 4=1
	.dw	i_4eq4			; 24 4=4
	.dw	i_4eq0			; 25 4=0
	.dw	i_4be0			; 26 4<0
	.dw	i_4ae0			; 27 4>=0
	.dw	i_cfetch		; 28 c@
	.dw	i_cstore		; 29 c!
	.dw	i_cfetchplus		; 2a c@+
	.dw	i_cstoreplus		; 2b c!+
	.dw	i_fetch			; 2c @
	.dw	i_store			; 2d !
	.dw	i_to_r			; 2e >r
	.dw	i_r_to			; 2f r>
	.dw	i_to_a			; 30 >a
	.dw	i_a_to			; 31 a>
	.dw	i_if			; 32 if
	.dw	i_if0			; 33 if0
	.dw	i_is			; 34 is
	.dw	i_is0			; 35 is0
	.dw	i_jump			; 36 jump
	.dw	i_jumpf			; 37 jumpf
	.dw	i_jump8			; 38 jump8
	.dw	i_call			; 39 call
	.dw	i_callf			; 3a callf
	.dw	i_call8			; 3b call8
	.dw	i_ret			; 3c ret
	.dw	i_retf			; 3d retf
	.dw	i_lit			; 3e lit
	.dw	i_4lit			; 3f 4lit
	.dw	i_alit			; 40 alit
	.dw	i_ialit			; 41 ialit
	.dw	i_sys			; 42 sys


reset:
	ldi	r16,low(ramend)
	ldi	r17,high(ramend)
	out	spl,r16
	out	sph,r17			; Set hardware stack pointer.
	ldi	apl,low(ramend-ssize)
	ldi	aph,high(ramend-ssize)
	ldi	rpl,low(ramend-ssize-asize)
	ldi	rph,high(ramend-ssize-asize)
	ldi	ppl,low(ramend-ssize-asize-rsize)
	ldi	pph,high(ramend-ssize-asize-rsize)

	rcall	spi_init

	ser	r16
	sts	buffer1l,r16
	sts	buffer1h,r16
	sts	buffer2l,r16
	sts	buffer2h,r16
	sts	bufferrecent,r16
	clr	r16
	sts	bufferstat,r16

	clr	r4
	clr	r5
	clr	r6
	ldi	r16,0x01
	mov	r7,r16

	clr	r8
	clr	r9
	clr	r10
	clr	r11


; execute
;
; The main execution loop. You can't escape from it!

execute:
	mov	r12,r4
	mov	r13,r5
	mov	r14,r6
	mov	r15,r7
	add	r12,r8
	adc	r13,r9
	adc	r14,r10
	rcall	read_byte		; Read instruction 
	mov	zl,r16
	lsl	zl			; z = z * 2
	clr	zh
	adiw	zl,instruction_table*2	; z += instruction_table
	lpm
	mov	r1,r0			; r1 = program[z]
	adiw	zl,1			; z++
	lpm				; r0 = program[z]
	mov	zl,r1
	mov	zh,r0			; z = r1:r0
	ldi	r16,1
	rcall	ip_add
	icall				; call z
	rjmp	execute			; execute next instruction


; spi_init
;
; Initialize SPI, so that we can play with the DataFlash.

spi_init:
	ldi	r16,0x10
	out	portb,r16
	ldi	r16,0xb0
	out	ddrb,r16
	ldi	r16,0x5c
	out	spcr,r16
	ret


; spi_rw
;
; Write the byte in r16, then read a byte to r16.

spi_rw:
	out	spdr,r16
spi_rw_wait:
	sbis	spsr,7
	rjmp	spi_rw_wait
	in	r16,spdr
	ret


; flash_read_status
;
; Read status byte to r16.

flash_read_status:
	sbi	portb,4
	cbi	portb,4
	ldi	r16,0x57
	rcall	spi_rw
	clr	r16
	rcall	spi_rw
	ret


; flash_page_to_buffer
;
; r18:r17 = page to load
; r19 = 0 (buffer 1) or 1 (buffer 2).

flash_page_to_buffer:
	sbi	portb,4
	cbi	portb,4
	clc
	rol	r17
	rol	r18
	lsl	r19
	ldi	r16,0x53
	add	r16,r19
	rcall	spi_rw
	mov	r16,r18
	rcall	spi_rw
	mov	r16,r17
	rcall	spi_rw
	clr	r16
	rcall	spi_rw
	sbi	portb,4
	cbi	portb,4
flash_page_to_buffer_wait:
	rcall	flash_read_status
	sbrs	r16,7
	rjmp	flash_page_to_buffer_wait
	ret


; flash_read_byte

; r18:r17 = offset within buffer
; r19 = 0 (buffer 1) or 1 (buffer 2).
; Data byte returned in r16

flash_read_byte:
	sbi	portb,4
	cbi	portb,4
	lsl	r19
	ldi	r16,0x54
	add	r16,r19
	rcall	spi_rw
	clr	r16
	rcall	spi_rw
	mov	r16,r18
	rcall	spi_rw
	mov	r16,r17
	rcall	spi_rw
	clr	r16
	rcall	spi_rw
	clr	r16
	rjmp	spi_rw



; flash_buffer_to_page
;
; r18:r17 = page to save
; r19 = 0 (buffer 1) or 1 (buffer 2).

flash_buffer_to_page:
	sbi	portb,4
	cbi	portb,4
	clc
	rol	r17
	rol	r18
	cpi	r19,1
	breq	flash_buffer_to_page_use2
flash_buffer_to_page_use1:
	ldi	r16,0x83
	rjmp	flash_buffer_to_page_go
flash_buffer_to_page_use2:
	ldi	r16,0x86
flash_buffer_to_page_go:
	rcall	spi_rw
	mov	r16,r18
	rcall	spi_rw
	mov	r16,r17
	rcall	spi_rw
	clr	r16
	rcall	spi_rw
	sbi	portb,4
	cbi	portb,4
flash_buffer_to_page_wait:
	rcall	flash_read_status
	sbrs	r16,7
	rjmp	flash_buffer_to_page_wait
	ret


; flash_save1
;
; Write buffer 1 to flash if cache modified.

flash_save1:
	lds	r16,bufferstat
	sbrs	r16,buffer1dirty
	ret

	andi	r16,0xfe
	sts	bufferstat,r16
	lds	r17,buffer1l
	lds	r18,buffer1h
	clr	r19
	rjmp	flash_buffer_to_page


; flash_save2
;
; Write buffer 2 to flash if cache modified.

flash_save2:
	lds	r16,bufferstat
	sbrs	r16,buffer2dirty
	ret

	andi	r16,0xfd
	sts	bufferstat,r16
	lds	r17,buffer2l
	lds	r18,buffer2h
	ldi	r19,1
	rjmp	flash_buffer_to_page



; flash_read
;
; r22:r21:r20 = byte to read, r22:r21 is the page number, r20 the offset.
; Byte returned in r16.

flash_read:
	lds	r16,buffer1l
	cp	r16,r21
	brne	flash_read_try2
	lds	r16,buffer1h
	cp	r16,r22
	brne	flash_read_try2
flash_read_load1:
	rjmp	flash_read_use1b
flash_read_use1:
	rcall	flash_save1
	clr	r19
	mov	r17,r21
	mov	r18,r22
	sts	buffer1l,r21
	sts	buffer1h,r22
	rcall	flash_page_to_buffer
flash_read_use1b:
	clr	r16
	sts	bufferrecent,r16
	mov	r17,r20
	clr	r18
	clr	r19
	rjmp	flash_read_byte
flash_read_try2:
	lds	r16,buffer2l
	cp	r16,r21
	brne	flash_read_new
	lds	r16,buffer2h
	cp	r16,r22
	brne	flash_read_new
flash_read_load2:
	rjmp	flash_read_use2b
flash_read_use2:
	rcall	flash_save2
	ldi	r19,1
	mov	r17,r21
	mov	r18,r22
	sts	buffer2l,r21
	sts	buffer2h,r22
	rcall	flash_page_to_buffer
flash_read_use2b:
	ldi	r16,1
	sts	bufferrecent,r16
	mov	r17,r20
	clr	r18
	ldi	r19,1
	rjmp	flash_read_byte
flash_read_new:
	lds	r16,bufferrecent
	sbrs	r16,0
	rjmp	flash_read_use2
	rjmp	flash_read_use1


; flash_write_byte
;
; r16 = buffer to use, r18:r17 = offset, r23 = data byte.

flash_write_byte:
	sbi	portb,4
	cbi	portb,4
	cpi	r16,1
	breq	flash_write_byte_use2
flash_write_byte_use1:
	lds	r16,bufferstat
	ori	r16,1
	sts	bufferstat,r16
	ldi	r16,0x84
	rjmp	flash_write_byte_done
flash_write_byte_use2:
	lds	r16,bufferstat
	ori	r16,2
	sts	bufferstat,r16
	ldi	r16,0x87
flash_write_byte_done:
	rcall	spi_rw
	clr	r16
	rcall	spi_rw
	mov	r16,r18
	rcall	spi_rw
	mov	r16,r17
	rcall	spi_rw
	mov	r16,r23
	rjmp	spi_rw


; flash_write
;
; r22:r21:r20 = address to write, r22:r21 is the page number, r20 the offset.
; Data byte in r23.

flash_write:
	lds	r16,buffer1l
	cp	r16,r21
	brne	flash_write_try2
	lds	r16,buffer1h
	cp	r16,r22
	brne	flash_write_try2
flash_write_load1:
	rjmp	flash_write_use1b
flash_write_use1:
	rcall	flash_save1
	clr	r19
	mov	r17,r21
	mov	r18,r22
	sts	buffer1l,r21
	sts	buffer1h,r22
	rcall	flash_page_to_buffer
flash_write_use1b:
	clr	r16
	sts	bufferrecent,r16
	mov	r17,r20
	clr	r18
	rjmp	flash_write_byte
flash_write_try2:
	lds	r16,buffer2l
	cp	r16,r21
	brne	flash_write_new
	lds	r16,buffer2h
	cp	r16,r22
	brne	flash_write_new
flash_write_load2:
	rjmp	flash_write_use2b
flash_write_use2:
	rcall	flash_save2
	ldi	r19,1
	mov	r17,r21
	mov	r18,r22
	sts	buffer2l,r21
	sts	buffer2h,r22
	rcall	flash_page_to_buffer
flash_write_use2b:
	ldi	r16,1
	sts	bufferrecent,r16
	mov	r17,r20
	clr	r18
	rjmp	flash_write_byte
flash_write_new:
	lds	r16,bufferrecent
	sbrs	r16,0
	rjmp	flash_write_use2
	rjmp	flash_write_use1



; read_eeprom
;
; Read one byte from the internal EEPROM at r12:r13 to r0

read_eeprom:
	sbic	eecr,eewe		; Is EEPROM ready?
	rjmp	read_eeprom		; No, wait.
	out	eearl,r12
	out	eearh,r13		; Output address.
	sbi	eecr,eere		; Execute read command.
	in	r0,eedr			; Read data byte.
	ret


; write_eeprom
;
; Write one byte from r16 to the internal EEPROM at r12:r13.

write_eeprom:
	sbic	eecr,eewe		; Ready?
	rjmp	write_eeprom		; Nope, wait.
	out	eearl,r12
	out	eearh,r13		; output address
	out	eedr,r16		; output data
	sbi	eecr,eemwe
	sbi	eecr,eewe		; write
	ret


; read_byte
;
; Read one byte from the memory address in r12:r15 to r16

read_byte:
	mov	r16,r15
	cpi	r16,0x00
	breq	read_byte_ram
	cpi	r16,0x01
	breq	read_byte_eeprom
	cpi	r16,0x02
	breq	read_byte_rom
	cpi	r16,0x03
	breq	read_byte_flash
	ldi	r16,0x00		; No, just load a zero.
	ret
read_byte_ram:
	mov	zl,r12
	mov	zh,r13			; Move address to Z.
	ld	r16,z			; Load to r16.
	ret
read_byte_eeprom:
	rcall	read_eeprom
	mov	r16,r0
	ret
read_byte_rom:
	mov	zl,r12
	mov	zh,r13			; Move address to Z.
	lpm
	mov	r16,r0			; Load to r16.
	ret
read_byte_flash:
	mov	r20,r12
	mov	r21,r13
	mov	r22,r14
	rjmp	flash_read


add_flash_address:
	ldi	r16,1
	clr	r17
	add	r20,r16
	adc	r21,r17
	adc	r22,r17
	ret


; read_word
;
; Read one word from the memory address in r12:r15 to r16:r19

read_word:
	mov	r16,r15
	cpi	r16,0x00
	breq	read_word_ram
	cpi	r16,0x01
	breq	read_word_eeprom
	cpi	r16,0x02
	breq	read_word_rom
	cpi	r16,0x03
	breq	read_word_flash
	clr	r16
	clr	r17
	clr	r18
	clr	r19
	ret
read_word_ram:
	mov	zl,r12
	mov	zh,r13			; Move address to Z.
	ld	r16,z+			; Load to r16.
	ld	r17,z+			; Load to r16.
	ld	r18,z+			; Load to r16.
	ld	r19,z+			; Load to r16.
	ret
read_word_eeprom:
	ldi	r20,1
	clr	r21
	rcall	read_eeprom
	mov	r16,r0
	add	r12,r20
	adc	r13,r21
	rcall	read_eeprom
	mov	r17,r0
	add	r12,r20
	adc	r13,r21
	rcall	read_eeprom
	mov	r18,r0
	add	r12,r20
	adc	r13,r21
	rcall	read_eeprom
	mov	r19,r0
	add	r12,r20
	adc	r13,r21
	ret
read_word_rom:
	mov	zl,r12
	mov	zh,r13			; Move address to Z.
	lpm
	mov	r16,r0			; Load to r16.
	adiw	zl,1
	lpm
	mov	r17,r0			; Load to r17.
	adiw	zl,1
	lpm
	mov	r18,r0			; Load to r18.
	adiw	zl,1
	lpm
	mov	r19,r0			; Load to r19.
	ret
read_word_flash:
	mov	r20,r12
	mov	r21,r13
	mov	r22,r14
	rcall	flash_read
	push	r16
	rcall	add_flash_address
	rcall	flash_read
	push	r16
	rcall	add_flash_address
	rcall	flash_read
	push	r16
	rcall	add_flash_address
	rcall	flash_read
	mov	r19,r16
	pop	r18
	pop	r17
	pop	r16
	ret


; write_byte
;
; Write one byte to the memory address in r12:r15 from r16

write_byte:
	mov	r17,r15
	cpi	r17,0x00
	breq	write_byte_ram
	cpi	r17,0x01
	breq	write_byte_eeprom
	cpi	r17,0x03
	breq	write_byte_flash
	ret				; No, do nothing.
write_byte_ram:
	mov	zl,r12
	mov	zh,r13			; Move address to Z.
	st	z,r16			; Store from r16
	ret
write_byte_eeprom:
	rjmp	write_eeprom
write_byte_flash:
	mov	r20,r12
	mov	r21,r13
	mov	r22,r14
	mov	r23,r16
	rjmp	flash_write


; write_word
;
; Write one word to the memory address in r12:r15 from r16:r19

write_word:
	mov	r20,r15
	cpi	r20,0x00
	breq	write_word_ram
	cpi	r20,0x01
	breq	write_word_eeprom
	cpi	r20,0x03
	breq	write_word_flash
	ret				; No, do nothing.
write_word_ram:
	mov	zl,r12
	mov	zh,r13			; Move address to Z.
	st	z+,r16			; Store from r16
	st	z+,r17			; Store from r17
	st	z+,r18			; Store from r18
	st	z+,r19			; Store from r19
	ret
write_word_eeprom:
	ldi	r20,1
	clr	r21
	rcall	write_eeprom
	add	r12,r20
	adc	r13,r21
	mov	r16,r17
	rcall	write_eeprom
	add	r12,r20
	adc	r13,r21
	mov	r16,r18
	rcall	write_eeprom
	add	r12,r20
	adc	r13,r21
	mov	r16,r19
	rjmp	write_eeprom
write_word_flash:
	push	r19
	push	r18
	push	r17
	mov	r20,r12
	mov	r21,r13
	mov	r22,r14
	mov	r23,r16
	rcall	flash_write
	rcall	add_flash_address
	pop	r23
	rcall	flash_write
	rcall	add_flash_address
	pop	r23
	rcall	flash_write
	rcall	add_flash_address
	pop	r23
	rjmp	flash_write



; ip_add
;
; Add the value of r16 to r12:r14

ip_add:
	clr	r17
	add	r8,r16
	adc	r9,r17
	adc	r10,r17
	ret

ip_load:
	mov	r12,r4
	mov	r13,r5
	mov	r14,r6
	mov	r15,r7
	add	r12,r8
	adc	r13,r9
	adc	r14,r10
	ret


i_drop:
	adiw	ppl,1
	ret

i_dup:
	ld	r16,y
	st	-y,r16
	ret

i_over:
	ldd	r16,y+1
	st	-y,r16
	ret

i_swap:
	ld	r16,y+
	ld	r17,y
	st	y,r16
	st	-y,r17
	ret

i_4drop:
	adiw	ppl,4
	ret

i_4dup:
	ldd	r16,y+3
	st	-y,r16
	ldd	r16,y+3
	st	-y,r16
	ldd	r16,y+3
	st	-y,r16
	ldd	r16,y+3
	st	-y,r16
	ret

i_4over:
	ldd	r16,y+7
	st	-y,r16
	ldd	r16,y+7
	st	-y,r16
	ldd	r16,y+7
	st	-y,r16
	ldd	r16,y+7
	st	-y,r16
	ret

i_4swap:
	ld	r16,y
	ldd	r17,y+4
	std	y+4,r16
	st	y,r17
	ldd	r16,y+1
	ldd	r17,y+5
	std	y+5,r16
	std	y+1,r17
	ldd	r16,y+2
	ldd	r17,y+6
	std	y+6,r16
	std	y+2,r17
	ldd	r16,y+3
	ldd	r17,y+7
	std	y+7,r16
	std	y+3,r17
	ret

i_adrop:
	adiw	apl,4
	ret

i_adup:
	mov	zl,apl
	mov	zh,aph
	ldd	r16,z+3
	st	-z,r16
	ldd	r16,z+3
	st	-z,r16
	ldd	r16,z+3
	st	-z,r16
	ldd	r16,z+3
	st	-z,r16
	sbiw	apl,4
	ret

i_aover:
	mov	zl,apl
	mov	zh,aph
	ldd	r16,z+7
	st	-z,r16
	ldd	r16,z+7
	st	-z,r16
	ldd	r16,z+7
	st	-z,r16
	ldd	r16,z+7
	st	-z,r16
	sbiw	apl,4
	ret

i_aswap:
	mov	zl,apl
	mov	zh,aph
	ld	r16,z
	ldd	r17,z+4
	std	z+4,r16
	st	z,r17
	ldd	r16,z+1
	ldd	r17,z+5
	std	z+5,r16
	std	z+1,r17
	ldd	r16,z+2
	ldd	r17,z+6
	std	z+6,r16
	std	z+2,r17
	ldd	r16,z+3
	ldd	r17,z+7
	std	z+7,r16
	std	z+3,r17
	ret

i_plus:
	ld	r16,y+			; fetch TOS
	ld	r17,y			; fetch NOS
	add	r17,r16			; add the bytes
	st	y,r17
	ret

i_minus:
	ld	r16,y+			; fetch TOS
	ld	r17,y			; fetch NOS
	sub	r17,r16			; subtract
	st	y,r17
	ret

i_and:
	ld	r16,y+			; fetch TOS
	ld	r17,y			; fetch NOS
	and	r17,r16
	st	y,r17
	ret

i_or:
	ld	r16,y+			; fetch TOS
	ld	r17,y			; fetch NOS
	or	r17,r16
	st	y,r17
	ret

i_xor:
	ld	r16,y+			; fetch TOS
	ld	r17,y			; fetch NOS
	eor	r17,r16
	st	y,r17
	ret

i_neg:
	ld	r16,y			; fetch TOS
	clr	r17
	sub	r17,r16			; r17 = 0-r16
	st	y,r17
	ret

i_not:
	ld	r16,y			; fetch TOS
	com	r16			; TOS = TOS xor 0xff
	st	y,r16
	ret

i_mul2:
	ld	r16,y			; fetch TOS
	clc				; clear carry flag
	rol	r16			; shift left 1 step
	st	y,r16			; store TOS
	ret

i_div2:
	ld	r16,y			; fetch TOS
	clc				; clear carry flag
	ror	r16			; shift right 1 step
	st	y,r16			; store TOS
	ret

i_4plus1:
	ld	r16,y+
	ld	r17,y
	add	r17,r16
	st	y,r17
	brcc	i_4plus1_done
	ldi	r16,0
	ldd	r17,y+1
	adc	r17,r16
	std	y+1,r17
	brcc	i_4plus1_done
	ldd	r17,y+2
	adc	r17,r16
	std	y+2,r17
	brcc	i_4plus1_done
	ldd	r17,y+3
	adc	r17,r16
	std	y+3,r17
i_4plus1_done:
	ret

i_4minus1:
	ld	r16,y+
	ld	r17,y
	sub	r17,r16
	st	y,r17
	brcc	i_4minus1_done
	ldi	r16,0
	ldd	r17,y+1
	sbc	r17,r16
	std	y+1,r17
	brcc	i_4minus1_done
	ldd	r17,y+2
	sbc	r17,r16
	std	y+2,r17
	brcc	i_4minus1_done
	ldd	r17,y+3
	sbc	r17,r16
	std	y+3,r17
i_4minus1_done:
	ret

i_4neg:
	clr	r16
	clr	r17
	clr	r18
	clr	r19
	ld	r20,y
	sub	r16,r20
	st	y,r16
	ldd	r20,y+1
	sub	r17,r20
	std	y+1,r17
	ldd	r20,y+2
	sub	r18,r20
	std	y+2,r18
	ldd	r20,y+3
	sub	r19,r20
	std	y+3,r19
	ret

i_4shl:
	ld	r20,y+
	ld	r16,y
	ldd	r17,y+1
	ldd	r18,y+2
	ldd	r19,y+3
i_4shl_next:
	clc
	rol	r16
	rol	r17
	rol	r18
	rol	r19
	dec	r20
	brne	i_4shl_next

	st	y,r16
	std	y+1,r17
	std	y+2,r18
	std	y+3,r19
	ret

i_4shr:
	ld	r20,y+
	ld	r16,y
	ldd	r17,y+1
	ldd	r18,y+2
	ldd	r19,y+3
i_4shr_next:
	clc
	ror	r19
	ror	r18
	ror	r17
	ror	r16
	dec	r20
	brne	i_4shr_next

	st	y,r16
	std	y+1,r17
	std	y+2,r18
	std	y+3,r19
	ret

i_4plus4:
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y+
	ld	r20,y
	add	r20,r16
	st	y,r20
	ldd	r20,y+1
	adc	r20,r17
	std	y+1,r20
	ldd	r20,y+2
	adc	r20,r18
	std	y+2,r20
	ldd	r20,y+3
	adc	r20,r19
	std	y+3,r20
	ret

i_4minus4:
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y+
	ld	r20,y
	sub	r20,r16
	st	y,r20
	ldd	r20,y+1
	sbc	r20,r17
	std	y+1,r20
	ldd	r20,y+2
	sbc	r20,r18
	std	y+2,r20
	ldd	r20,y+3
	sbc	r20,r19
	std	y+3,r20
	ret

i_4and4:
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y+
	ld	r20,y
	and	r20,r16
	st	y,r20
	ldd	r20,y+1
	and	r20,r17
	std	y+1,r20
	ldd	r20,y+2
	and	r20,r18
	std	y+2,r20
	ldd	r20,y+3
	and	r20,r19
	std	y+3,r20
	ret

i_4or4:
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y+
	ld	r20,y
	or	r20,r16
	st	y,r20
	ldd	r20,y+1
	or	r20,r17
	std	y+1,r20
	ldd	r20,y+2
	or	r20,r18
	std	y+2,r20
	ldd	r20,y+3
	or	r20,r19
	std	y+3,r20
	ret

i_4xor4:
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y+
	ld	r20,y
	eor	r20,r16
	st	y,r20
	ldd	r20,y+1
	eor	r20,r17
	std	y+1,r20
	ldd	r20,y+2
	eor	r20,r18
	std	y+2,r20
	ldd	r20,y+3
	eor	r20,r19
	std	y+3,r20
	ret

i_eq:
	ld	r16,y+
	ld	r17,y
	cp	r16,r17
	breq	i_eq_true
	clr	r16
	st	y,r16
	ret
i_eq_true:
	ser	r16
	st	y,r16
	ret

i_neq:
	ld	r16,y+
	ld	r17,y
	sub	r16,r17
	brne	i_neq_true
	st	y,r16
	ret
i_neq_true:
	ser	r16
	st	y,r16
	ret

i_be:
	ld	r16,y+
	ld	r17,y
	sub	r16,r17
	brlo	i_be_true
	clr	r16
	st	y,r16
	ret
i_be_true:
	ser	r16
	st	y,r16
	ret

i_ab:
	ld	r16,y+
	ld	r17,y
	cp	r16,r17
	breq	i_ab_false
	brlo	i_ab_false
	ser	r16
	st	y,r16
	ret
i_ab_false:
	clr	r16
	st	y,r16
	ret

i_4eq1:
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y+
	ld	r20,y
	or	r17,r18
	or	r17,r19
	brne	i_4eq1_false
	cp	r16,r20
	breq	i_4eq1_true
i_4eq1_false:
	clr	r16
	st	y,r16
	ret
i_4eq1_true:
	ser	r16
	st	y,r16
	ret

i_4eq4:
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y+
	ld	r20,y+
	ld	r21,y+
	ld	r22,y+
	ld	r23,y
	sub	r16,r20
	sub	r17,r21
	sub	r18,r22
	sub	r19,r23
	or	r16,r17
	or	r16,r18
	or	r16,r19
	breq	i_4eq4_true
	clr	r16
	st	y,r16
	ret
i_4eq4_true:
	ser	r16
	st	y,r16
	ret

i_4eq0:
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y
	or	r16,r17
	or	r16,r18
	or	r16,r19
	breq	i_4eq0_true
	clr	r16
	st	y,r16
	ret
i_4eq0_true:
	ser	r16
	st	y,r16
	ret

i_4be0:
	adiw	yl,3
	ld	r16,y
	sbrc	r16,7
	rjmp	i_4be0_true
	clr	r16
	st	y,r16
	ret
i_4be0_true:
	ser	r16
	st	y,r16
	ret

i_4ae0:
	adiw	yl,3
	ld	r16,y
	sbrs	r16,7
	rjmp	i_4ae0_true
	clr	r16
	st	y,r16
	ret
i_4ae0_true:
	ser	r16
	st	y,r16
	ret

i_cfetch:
	mov	zl,apl
	mov	zh,aph
	ld	r12,z
	ldd	r13,z+1
	ldd	r14,z+2
	ldd	r15,z+3
	rcall	read_byte
	st	-y,r16
	ret

i_cstore:
	mov	zl,apl
	mov	zh,aph
	ld	r12,z
	ldd	r13,z+1
	ldd	r14,z+2
	ldd	r15,z+3
	ld	r16,y+
	rjmp	write_byte

i_cfetchplus:
	rcall	i_cfetch
i_aplus:
	ldi	r16,1
	clr	r17
	add	r12,r16
	adc	r13,r17
	adc	r14,r17
	mov	zl,apl
	mov	zh,aph
	st	z,r12
	std	z+1,r13
	std	z+2,r14
	ret

i_cstoreplus:
	rcall	i_cstore
	rjmp	i_aplus

i_fetch:
	mov	zl,apl
	mov	zh,aph
	ld	r12,z
	ldd	r13,z+1
	ldd	r14,z+2
	ldd	r15,z+3
	rcall	read_word
	st	-y,r19
	st	-y,r18
	st	-y,r17
	st	-y,r16
	ret

i_store:
	mov	zl,apl
	mov	zh,aph
	ld	r12,z
	ldd	r13,z+1
	ldd	r14,z+2
	ldd	r15,z+3
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y+
	rjmp	write_word

i_to_r:
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y+
	st	-x,r19
	st	-x,r18
	st	-x,r17
	st	-x,r16
	ret

i_r_to:
	ld	r16,x+
	ld	r17,x+
	ld	r18,x+
	ld	r19,x+
	st	-y,r19
	st	-y,r18
	st	-y,r17
	st	-y,r16
	ret

i_to_a:
	ld	r16,y+
	ld	r17,y+
	ld	r18,y+
	ld	r19,y+
	mov	zl,apl
	mov	zh,aph
	st	-z,r19
	st	-z,r18
	st	-z,r17
	st	-z,r16
	sbiw	apl,4
	ret

i_a_to:
	mov	zl,apl
	mov	zh,aph
	ld	r16,z+
	ld	r17,z+
	ld	r18,z+
	ld	r19,z+
	adiw	apl,4
	st	-y,r19
	st	-y,r18
	st	-y,r17
	st	-y,r16
	ret

i_jump:
	rcall	ip_load
	rcall	read_word
	mov	r8,r16
	mov	r9,r17
	mov	r10,r18
	mov	r11,r19
	ret

i_jumpf:
	rcall	ip_load
	rcall	read_word
	mov	r8,r16
	mov	r9,r17
	mov	r10,r18
	mov	r11,r19
	ld	r4,y+
	ld	r5,y+
	ld	r6,y+
	ld	r7,y+
	ret

i_jump8:
	rcall	ip_load
	ldi	r16,1
	rcall	ip_add
	rcall	read_byte
	clr	r17
	sbrc	r16,7
	com	r17
	add	r8,r16
	adc	r9,r17
	adc	r10,r17
	adc	r11,r17
	ret

i_if:
	ld	r16,y+
	cpi	r16,0x00
	brne	i_jump8
	ldi	r16,1
	rjmp	ip_add

i_if0:
	ld	r16,y+
	cpi	r16,0x00
	breq	i_jump8
	ldi	r16,1
	rjmp	ip_add

i_is:
	ld	r16,y
	cpi	r16,0x00
	brne	i_jump8
	ldi	r16,1
	rjmp	ip_add

i_is0:
	ld	r16,y
	cpi	r16,0x00
	breq	i_jump8
	ldi	r16,1
	rjmp	ip_add

i_call:
	rcall	ip_load
	ldi	r16,4
	rcall	ip_add
	st	-x,r11
	st	-x,r10
	st	-x,r9
	st	-x,r8
	rcall	read_word
	mov	r8,r16
	mov	r9,r17
	mov	r10,r18
	mov	r11,r19
	ret

i_callf:
	rcall	ip_load
	ldi	r16,4
	rcall	ip_add
	st	-x,r7
	st	-x,r6
	st	-x,r5
	st	-x,r4
	st	-x,r11
	st	-x,r10
	st	-x,r9
	st	-x,r8
	rcall	read_word
	mov	r8,r16
	mov	r9,r17
	mov	r10,r18
	mov	r11,r19
	ld	r4,y+
	ld	r5,y+
	ld	r6,y+
	ld	r7,y+
	ret

i_call8:
	rcall	ip_load
	ldi	r16,1
	rcall	ip_add
	st	-x,r11
	st	-x,r10
	st	-x,r9
	st	-x,r8
	rcall	read_byte
	clr	r17
	sbrc	r16,7
	com	r17
	add	r8,r16
	adc	r9,r17
	adc	r10,r17
	adc	r11,r17
	ret

i_ret:
	ld	r8,x+
	ld	r9,x+
	ld	r10,x+
	ld	r11,x+
	ret

i_retf:
	rcall	i_ret
	ld	r4,x+
	ld	r5,x+
	ld	r6,x+
	ld	r7,x+
	ret

i_lit:
	rcall	ip_load
	rcall	read_byte
	st	-y,r16
	ldi	r16,1
	rjmp	ip_add

i_4lit:
	rcall	ip_load
	rcall	read_word
	st	-y,r19
	st	-y,r18
	st	-y,r17
	st	-y,r16
	ldi	r16,4
	rjmp	ip_add

i_alit:
	rcall	ip_load
	rcall	read_word
i_alit_push:
	mov	zl,apl
	mov	zh,aph
	st	-z,r19
	st	-z,r18
	st	-z,r17
	st	-z,r16
	sbiw	apl,4
	ldi	r16,4
	rjmp	ip_add

i_ialit:
	rcall	ip_load
	rcall	read_word
	add	r16,r4
	adc	r17,r5
	adc	r18,r6
	adc	r19,r7
	rjmp	i_alit_push

i_sys:
	ld	zl,y+			; pop z
	lsl	zl			;
	clr	zh			;
	adiw	zl,(syscall_table*2)	; z += syscall_table
	lpm				;
	mov	r1,r0			; r1 = program[z]
	adiw	zl,1			; z++
	lpm				; r0 = program[z]
	mov	zl,r1			;
	mov	zh,r0			; z = r1:r0
	icall				; call z
	ret


sys_uart_init:
	ld	r16,y+			; pop tmpa = baud rate
	out	ubrr,r16		; set baud rate
	ldi	r16,$18			; enable receive & transmit
	out	ucr,r16
	ret

sys_uart_poll:
	sbis	usr,rxc			; Receive complete?
	rjmp	sys_uart_poll_false
	ser	r16
	st	-y,r16
	ret
sys_uart_poll_false:
	clr	r16
	st	-y,r16
	ret

sys_uart_read:
	sbis	usr,rxc			; Receive complete?
	rjmp	sys_uart_read		; Nope, wait.
	in	r16,udr			; Input the data byte.
	st	-y,r16			; Push the byte.
	ret

sys_uart_write:
	sbis	usr,udre		; Previous write finished?
	rjmp	sys_uart_write		; Nope, wait.
	ld	r16,y+			; Pop the data byte.
	out	udr,r16			; Output the data byte.
	ret

sys_get_freeram:
	clr	r16
	st	-y,r16
	st	-y,r16
	ldi	r16,low(freeramend-freeramstart)
	ldi	r17,high(freeramend-freeramstart)
	st	-y,r17
	st	-y,r16
	ret

sys_sync_flash:
	rcall	flash_save1
	rjmp	flash_save2


.eseg
